#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define DEFAULT_FILENAME "sudoku.txt"

typedef struct {
	int x;
	int y;
} Coordinate;

typedef struct {
	int x;
	int y;
	int digit;
	int numOtherChoices;
	int setState;
	int *previousState;
	int previousStateChoices;
	int *previousNeighbourState1;
	int *previousNeighbourState2;
	int **previousNeighbourState3;
} Move;

int dimension;
int rootDimension;
int bias;
int ***grid;
int **gridChoices;

Move *moveStack;
int moveStackTop;

int singletonQueueSize;
Coordinate *singletonQueue;
int singletonQueueHead;
int singletonQueueTail;

int allocated = 0;


// Move stack
Move *MoveStackPush(void)
{
	return &moveStack[moveStackTop++];
}

Move *MoveStackPop(void)
{
	return &moveStack[--moveStackTop];
}

int MoveStackEmpty(void)
{
	return moveStackTop == 0;
}

void MoveStackClear(void)
{
	moveStackTop = 0;
}


// Singleton queue
void SingletonQueueEnqueue(Coordinate c)
{
	singletonQueue[singletonQueueHead] = c;
	singletonQueueHead = (singletonQueueHead + 1) % singletonQueueSize;
}

Coordinate SingletonQueueDequeue(void)
{
	Coordinate c;

	c = singletonQueue[singletonQueueTail];
	singletonQueueTail = (singletonQueueTail + 1) % singletonQueueSize;
	return c;
}

int SingletonQueueEmpty(void)
{
	return singletonQueueHead == singletonQueueTail;
}

void SingletonQueueClear(void)
{
	singletonQueueHead = 0;
	singletonQueueTail = 0;
}


// Display grid
void PrintGrid(void)
{
	int i, j, k, n;
	char c;

	for (i=0; i<dimension; i++)
	{
		printf("  ");
		for (j=0; j<dimension; j++)
		{
			if (gridChoices[i][j] == 0)
				c = 'x';
			else if (gridChoices[i][j] == 1)
			{
				for (k=0; k<dimension; k++)
					if (grid[i][j][k])
						break;
				n = k + bias;
				if (n <= 9)
					c = '0' + n;
				else
					c = 'A' + (n - 10);
			}
			else if (gridChoices[i][j] == dimension)
				c = ' ';
			else
				c = '_';
			printf("%c", c);
		}
		printf("\n");
	}
	printf("\n");
}


// Constraint propagation
int PropagateConstraints(int x, int y, int digit, int initialState)
{
	int i, j, n, p, q;
	int numOtherChoices;
	int borderX, borderY;
	int valid = 1;
	Coordinate singleton;
	Move *m;

	if (!initialState)
	{
		m = MoveStackPush();
		m->x = x;
		m->y = y;
		m->digit = digit;
	}

	// Set the state of the square if necessary
	if (gridChoices[x][y] > 1)
	{
		if (!initialState)
		{
			numOtherChoices = 0;
			for (i=digit+1; i<dimension; i++)
				if (grid[x][y][i])
					numOtherChoices++;
			m->numOtherChoices = numOtherChoices;

			m->setState = 1;
			for (i=0; i<dimension; i++)
				m->previousState[i] = grid[x][y][i];
			m->previousStateChoices = gridChoices[x][y];
		}

		for (i=0; i<dimension; i++)
			if (i != digit)
				grid[x][y][i] = 0;
		gridChoices[x][y] = 1;
	}
	else if (!initialState)
	{
		m->numOtherChoices = 0;
		m->setState = 0;
	}

	// Save the states of the neighbouring squares
	borderX = (x / rootDimension) * rootDimension;
	borderY = (y / rootDimension) * rootDimension;
	if (!initialState)
	{
		p = 0;
		q = 0;
		for (i=0; i<dimension; i++)
		{
			if (i != y)
				m->previousNeighbourState1[p++] = grid[x][i][digit];
			if (i != x)
				m->previousNeighbourState2[q++] = grid[i][y][digit];
		}
		p = 0;
		for (i=borderX; i<borderX+rootDimension; i++)
		{
			if (i != x)
			{
				q = 0;
				for (j=borderY; j<borderY+rootDimension; j++)
				{
					if (j != y)
					{
						m->previousNeighbourState3[p][q] = grid[i][j][digit];
						q++;
					}
				}
				p++;
			}
		}
	}

	// Propagate the constraints to the neighbouring squares
	for (i=0; i<dimension; i++)
	{
		if (i != y && grid[x][i][digit])
		{
			grid[x][i][digit] = 0;
			n = --gridChoices[x][i];
			if (n == 1)
			{
				singleton.x = x;
				singleton.y = i;
				SingletonQueueEnqueue(singleton);
			}
			else if (n <= 0)
				valid = 0;
		}
		if (i != x && grid[i][y][digit])
		{
			grid[i][y][digit] = 0;
			n = --gridChoices[i][y];
			if (n == 1)
			{
				singleton.x = i;
				singleton.y = y;
				SingletonQueueEnqueue(singleton);
			}
			else if (n <= 0)
				valid = 0;
		}
	}
	for (i=borderX; i<borderX+rootDimension; i++)
	{
		if (i != x)
		{
			for (j=borderY; j<borderY+rootDimension; j++)
			{
				if (j != y && grid[i][j][digit])
				{
					grid[i][j][digit] = 0;
					n = --gridChoices[i][j];
					if (n == 1)
					{
						singleton.x = i;
						singleton.y = j;
						SingletonQueueEnqueue(singleton);
					}
					else if (n <= 0)
						valid = 0;
				}
			}
		}
	}

	return valid;
}

// Backtracking
int Backtrack(void)
{
	int i, j, p, q;
	int choicePoint;
	int borderX, borderY;
	Move *m;

	SingletonQueueClear();

	choicePoint = 0;
	while (!MoveStackEmpty())
	{
		m = MoveStackPop();
		// Restore the state of the square if necessary
		if (m->setState)
		{
			for (i=0; i<dimension; i++)
				grid[m->x][m->y][i] = m->previousState[i];
			gridChoices[m->x][m->y] = m->previousStateChoices;
		}

		// Restore the states of the neighbouring squares
		p = 0;
		q = 0;
		for (i=0; i<dimension; i++)
		{
			if (i != m->y)
			{
				if (m->previousNeighbourState1[p])
				{
					grid[m->x][i][m->digit] = 1;
					gridChoices[m->x][i]++;
				}
				p++;
			}
			if (i != m->x)
			{
				if (m->previousNeighbourState2[q])
				{
					grid[i][m->y][m->digit] = 1;
					gridChoices[i][m->y]++;
				}
				q++;
			}
		}
		borderX = (m->x / rootDimension) * rootDimension;
		borderY = (m->y / rootDimension) * rootDimension;
		p = 0;
		for (i=borderX; i<borderX+rootDimension; i++)
		{
			if (i != m->x)
			{
				q = 0;
				for (j=borderY; j<borderY+rootDimension; j++)
				{
					if (j != m->y)
					{
						if (m->previousNeighbourState3[p][q])
						{
							grid[i][j][m->digit] = 1;
							gridChoices[i][j]++;
						}
						q++;
					}
				}
				p++;
			}
		}

		// Stop if reached a choice point
		if (m->numOtherChoices)
		{
			choicePoint = 1;
			break;
		}
	}

	// Expand the next choice
	if (choicePoint)
	{
		for (i=m->digit+1; i<dimension; i++)
			if (grid[m->x][m->y][i])
				break;
printf("BBBB [%d,%d] --> %d\n", m->x+1, m->y+1, i+1);
		return PropagateConstraints(m->x, m->y, i, 0);
	}
	else
		return -1;
}

// Top-level call to solve grid
int SolveGrid(void)
{
	int i, x, y;
	int numSolutions = 0;
	int numChoices, bestChoices;
	int bestX, bestY;
	int backtrack;
	Coordinate c;

	while (1)
	{
		do
		{
			// Propagate all the singletons
			while (!SingletonQueueEmpty())
			{
				c = SingletonQueueDequeue();
				for (i=0; i<dimension; i++)
					if (grid[c.x][c.y][i])
						break;
printf("s    [%d,%d] --> %d\n", c.x+1, c.y+1, i+1);
				if (!PropagateConstraints(c.x, c.y, i, 0))
				{
					while (!(backtrack = Backtrack()));
					if (backtrack == -1)
						return numSolutions;
					break;
				}
			}

			// Choose the first square with the smallest number of choices
			bestChoices = 0;
			for (x=0; x<dimension; x++)
			{
				for (y=0; y<dimension; y++)
				{
					numChoices = gridChoices[x][y];
					if (numChoices >= 2 && (!bestChoices || numChoices < bestChoices))
					{
						bestChoices = numChoices;
						bestX = x;
						bestY = y;
					}
				}
			}
			if (bestChoices)
			{
				for (i=0; i<dimension; i++)
					if (grid[bestX][bestY][i])
						break;
printf("cc   [%d,%d] --> %d\n", bestX+1, bestY+1, i+1);
				if (!PropagateConstraints(bestX, bestY, i, 0))
				{
					while (!(backtrack = Backtrack()));
					if (backtrack == -1)
						return numSolutions;
				}
			}
		}
		while (bestChoices);

		numSolutions++;
printf("\n%d)\n", numSolutions);
		PrintGrid();

		while (!(backtrack = Backtrack()));
		if (backtrack == -1)
			return numSolutions;
	}
}


// Grid construction
int BuildGrid(FILE *ifp)
{
	int width = 0, height = 0;
	int i, j, k, n, ch;
	int lowerBound, upperBound;
	int eof = 0;
	int contradiction = 0;

	height = 0;
	lowerBound = -1;
	upperBound = -1;
	for (i=0; !eof; i++)
	{
		j = 0;
		ch = fgetc(ifp);
		while (ch != '\n' && ch != EOF)
		{
			if (ch != ' ')
			{
				if (ch >= '0' && ch <= '9')
					n = ch - '0';
				else if (ch >= 'A' && ch <= 'Z')
					n = 10 + (ch - 'A');
				else if (ch >= 'a' && ch <= 'z')
					n = 10 + (ch - 'a');
				else
					n = -1;
				if (n != -1)
				{
					if (lowerBound == -1 || n < lowerBound)
						lowerBound = n;
					if (upperBound == -1 || n > upperBound)
						upperBound = n;
				}
			}

			j++;
			ch = fgetc(ifp);
		}
		if (ch == EOF)
			eof = 1;

		if (j > width)
			width = j;
		if (!eof || j != 0)
			height++;
	}
	dimension = width > height ? width : height;
	rootDimension = (int)sqrt(dimension);
	if (rootDimension * rootDimension != dimension)
		return 2;
	if (lowerBound == 0 && upperBound < dimension)
		bias = 0;
	else if (lowerBound >= 1 && upperBound <= dimension)
		bias = 1;
	else
		bias = lowerBound;

	grid = malloc(dimension * sizeof(int**));
	gridChoices = malloc(dimension * sizeof(int*));
	for (i=0; i<dimension; i++)
	{
		grid[i] = malloc(dimension * sizeof(int*));
		gridChoices[i] = malloc(dimension * sizeof(int));
		for (j=0; j<dimension; j++)
		{
			grid[i][j] = malloc(dimension * sizeof(int));
			for (k=0; k<dimension; k++)
				grid[i][j][k] = 1;
			gridChoices[i][j] = dimension;
		}
	}
	moveStack = malloc(dimension * dimension * sizeof(Move));
	for (i=0; i<dimension * dimension; i++)
	{
		moveStack[i].previousState = malloc(dimension * sizeof(int));
		moveStack[i].previousNeighbourState1 = malloc((dimension - 1) *
				sizeof(int));
		moveStack[i].previousNeighbourState2 = malloc((dimension - 1) *
				sizeof(int));
		moveStack[i].previousNeighbourState3 = malloc((rootDimension - 1) *
				sizeof(int*));
		for (j=0; j<rootDimension-1; j++)
			moveStack[i].previousNeighbourState3[j] = malloc((rootDimension - 1) *
					sizeof(int));
	}
	MoveStackClear();
	singletonQueueSize = dimension * dimension + 1;
	singletonQueue = malloc(singletonQueueSize * sizeof(Coordinate));
	SingletonQueueClear();
	allocated = 1;

	fseek(ifp, 0, SEEK_SET);
	eof = 0;
	for (i=0; !eof; i++)
	{
		j=0;
		ch = fgetc(ifp);
		while (ch != '\n' && ch != EOF)
		{
			if (ch != ' ')
			{
				if (ch >= '0' && ch <= '9')
					n = ch - '0';
				else if (ch >= 'A' && ch <= 'Z')
					n = 10 + (ch - 'A');
				else if (ch >= 'a' && ch <= 'z')
					n = 10 + (ch - 'a');
				else
					return 3;
				if (n < bias || n >= dimension + bias)
					return 4;

				if (!PropagateConstraints(i, j, n - bias, 1))
					contradiction = 1;
			}

			j++;
			ch = fgetc(ifp);
		}
		if (ch == EOF)
			eof = 1;
	}

	return contradiction;
}

// Cleanup
void FreeGrid(void)
{
	int i, j;

	if (allocated)
	{
		for (i=0; i<dimension; i++)
		{
			for (j=0; j<dimension; j++)
				free(grid[i][j]);
			free(grid[i]);
			free(gridChoices[i]);
		}
		free(grid);
		free(gridChoices);
		for (i=0; i<dimension * dimension; i++)
		{
			free(moveStack[i].previousState);
			free(moveStack[i].previousNeighbourState1);
			free(moveStack[i].previousNeighbourState2);
			for (j=0; j<rootDimension-1; j++)
				free(moveStack[i].previousNeighbourState3[j]);
			free(moveStack[i].previousNeighbourState3);
		}
		free(moveStack);
		free(singletonQueue);
	}
}


void PrintUsage(void)
{
	printf("Usage: Sudoku.exe [filename]");
}


int main(int argc, char **argv)
{
	int error;
	int numSolutions;
	char *filename;
	FILE *ifp;

	if (argc < 2)
		filename = DEFAULT_FILENAME;
	else if (argc == 2)
		filename = argv[1];
	else
	{
		PrintUsage();
getchar();
		return 1;
	}

	ifp = fopen(filename, "rt");
	if (!ifp)
	{
		PrintUsage();
getchar();
		return 1;
	}
	error = BuildGrid(ifp);
	fclose(ifp);
	if (!dimension || error)
	{
		FreeGrid();
		if (error == 1)
			printf("Contradiction found in file \"%s\"\n", filename);
		else
		{
			printf("Error reading file \"%s\"\n", filename);
			if (!dimension)
				printf("Invalid dimension\n");
			else if (error == 2)
				printf("Invalid maximum dimension (%d) not square\n", dimension);
			else if (error == 3)
				printf("Invalid character in file\n");
			else if (error == 4)
				printf("Grid square specified as greater than dimension (%d)\n",
						dimension);
		}
getchar();
		return 1;
	}

	PrintGrid();

	numSolutions = SolveGrid();
printf("\n%d %s\n", numSolutions, numSolutions == 1 ? "solution" : "solutions");

	FreeGrid();

getchar();
	return 0;
}
